#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ts=4:sw=4:et:tw=88:nowrap

#    This file is part of hough, an image deskewing tool
#    Copyright © 2016-2020 Toby Thain <toby@telegraphics.com.au>
#    Copyright © 2020-2025 Joan Touzet <joant@atypical.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import traceback
from multiprocessing import set_start_method
from pathlib import Path
from typing import Annotated

from cyclopts import App, Parameter
from cyclopts.types import ExistingFile

from hough import (
    CommonArgs,
    analyse_files,
    histogram as histo,
    logit,
    rotate_files,
)


app = App(default_parameter=Parameter(negative=()))


@app.command()
@logit
def process(
    files: list[Path],
    common: CommonArgs | None = None,
    histogram: Annotated[
        bool, Parameter(help="Display result summary as histogram")
    ] = False,
) -> int:
    """Fully analyse and rotate one or more files."""
    resultspath = analyse_files(files, common)
    if histogram:
        if ret := histo(resultspath) != 0:
            return ret
    return rotate_files(resultspath, common)


@app.command()
@logit
def analyse(
    files: list[ExistingFile],
    common: CommonArgs | None = None,
    histogram: bool = False,
) -> int:
    """Analyse one or more files for deskewing.

    Parameters
    ----------
    files: list[ExistingFile]
        One or more files to analyse for deskewing.
    histogram: bool
        Display result summary as histogram after processing.
    """
    resultspath = analyse_files(files, common)
    if not resultspath:
        return 1
    if histogram:
        return histo(resultspath)
    return 0


@app.command()
# no @logit - never log this one
def histogram(
    results: ExistingFile,
) -> int:
    """Show a histogram of rotation angles from a previous analysis."""
    return histo(results)


@app.command()
@logit
def rotate(
    results: ExistingFile,
    common: CommonArgs | None = None,
):
    """Rotate one or more files that have previously been analysed.

    Parameters
    ----------
    results: Path
        Use the specified file for analysis results.
    """
    results = results.absolute()
    # TODO: check all files mentioned in results exist too?
    return rotate_files(results, common)


def main(*args, **kwargs):
    # needed for CUDA backend multiprocessing pool, forced for pytest
    set_start_method("spawn", force=True)
    # TODO: maybe clean up empty out paths on exit?
    try:
        sys.exit(app(*args, **kwargs))
    except Exception:
        print("Uncaught exception:", file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import math

import numpy as np
from pathlib import Path
import termplotlib as tpl

from loguru import logger

from .utils import load_csv


def _do_histogram(results: Path):
    # load the file
    data = load_csv(results)
    res = []
    for r in data:
        for image in r:
            res.append(abs(image[2]))
    if not res:
        print("\nNo angles found, so no histogram.")
        return

    counts, bin_edges = np.histogram(res, bins=20, range=(0, int(math.ceil(max(res)))))
    labels = [
        "{:.2f}° - {:.2f}°".format(bin_edges[k], bin_edges[k + 1])
        for k in range(len(bin_edges) - 1)
    ]
    fig = tpl.figure()
    fig.barh(
        counts,
        labels=labels,
        max_width=40,
        bar_width=1,
        show_vals=True,
        force_ascii=False,
    )

    print("\n=== Skew statistics ===")
    fig.show()

    print(f"Samples: {len(res)}")
    print(f"50th percentile: {np.percentile(res, 50):1.2f}°")
    print(f"90th percentile: {np.percentile(res, 90):1.2f}°")
    print(f"99th percentile: {np.percentile(res, 99):1.2f}°")


def histogram(results: Path):
    try:
        _do_histogram(results)
        return 0
    except Exception:
        import traceback

        logger.error(f"Exception in histogram process: \n{traceback.format_exc()}")
        return 1



import os
from multiprocessing import set_start_method
from pathlib import Path
from shutil import rmtree

import pytest


# should not be necessary, but can't hurt.
@pytest.fixture(scope="session", autouse=True)
def always_spawn():
    set_start_method("spawn", force=True)


# always start a test run with clean coverage data
@pytest.fixture(scope="session", autouse=True)
def clean_sampledir():  # pragma: no cover
    for f in Path("samples").glob(".coverage.*"):
        f.unlink(missing_ok=True)
    yield


# Makes output cleaner when progress bar is drawn
@pytest.fixture(autouse=True)
def print_newline():
    print()
    yield
    print()


@pytest.fixture
def outpath():
    return "out"


# Run tests from samples directory, auto-cleaning up output directory
@pytest.fixture(autouse=True)
def sampledir(outpath):
    old_cwd = os.getcwd()
    os.chdir("samples")
    rmtree(outpath, ignore_errors=True)
    yield
    rmtree(outpath, ignore_errors=True)
    os.chdir(old_cwd)



from pathlib import Path

import pytest

from hough import (
    main,
)


def test_smask_pdf():
    with pytest.raises(SystemExit) as e:
        main(["process", "smask.pdf"])
    assert e.value.code == 0


def test_process_img():
    with pytest.raises(SystemExit) as e:
        main(["process", "deep_hurting.png"])
    assert e.value.code == 0


@pytest.mark.parametrize("outpath", ["t"])
def test_full_process():
    with pytest.raises(SystemExit) as e:
        main(
            [
                "process",
                "--histogram",
                "--out=t",
                "av-000.tif",
                "av-001.jpg",
            ]
        )
    assert e.value.code == 0
    assert Path("t").exists()
    assert Path("t/results.csv").exists()
    assert Path("t/av-000.tif").exists()
    # may not end up rotating
    # assert Path("t/av-001.jpg").exists()


# note: going to get DeprecationWarnings, cannot resolve yet
#   see https://github.com/swig/swig/issues/2881#issuecomment-2332652634



"""
Worker functions for a parallelizable skew analyser.
"""

import bisect
import csv
import math
from enum import Enum
from multiprocessing.pool import Pool
from pathlib import Path


try:  # pragma: no cover
    import cupy as cp
    import cupyx.scipy.ndimage as ndi
    from cucim.skimage.color import rgb2gray
    from cucim.skimage.exposure import is_low_contrast
    from cucim.skimage.feature import canny
    from cucim.skimage.filters import threshold_otsu
    from cucim.skimage.transform import downscale_local_mean
    from cucim.skimage.util import crop, img_as_bool, img_as_ubyte
except ModuleNotFoundError:  # pragma: no cover
    import numpy as cp
    import scipy.ndimage as ndi
    from skimage.color import rgb2gray
    from skimage.exposure import is_low_contrast
    from skimage.feature import canny
    from skimage.filters import threshold_otsu
    from skimage.transform import downscale_local_mean
    from skimage.util import crop, img_as_bool, img_as_ubyte

import imageio.v3 as iio
import numpy as np
from loguru import logger
from skimage.draw import line_aa
from skimage.transform import probabilistic_hough_line
from tqdm import tqdm  # TODO: consider rich?

from . import (
    CommonArgs,
    abort,
    get_images,
    get_num_images,
    get_cpu_image,
    get_gpu_image_maybe,
)


# TODO: configfile out all magic numbers, including windowsize
# TODO: class?

Orientation = Enum("Orientation", "HORIZONTAL,VERTICAL")


# hough's little helpers
def _greyf(arr):
    return (arr > 0.3) * 0.3


# CuPy scalars are actually GPU-resident zero-dimensional arrays, so we use numpy here instead
hough_prec = np.deg2rad(0.02)
hough_theta_h = np.arange(
    np.deg2rad(-93.0), np.deg2rad(-87.0), hough_prec, dtype=np.float64
)
hough_theta_v = np.arange(
    np.deg2rad(-3.0), np.deg2rad(3.0), hough_prec, dtype=np.float64
)
hough_theta_hv = np.concatenate((hough_theta_v, hough_theta_h))


def _hough_angles(pos, neg, orientation=Orientation.HORIZONTAL, thresh=(None, None)):
    height, width = pos.shape
    if orientation == Orientation.HORIZONTAL:
        axis = 1
        theta = hough_theta_h
        fp = cp.ones((51, 1), dtype=np.float32)
        margin = int(width * 0.975)
    elif orientation == Orientation.VERTICAL:
        axis = 0
        theta = hough_theta_v
        fp = cp.ones((1, 51), dtype=np.float32)
        margin = int(height * 0.025)
    else:  # pragma: no cover
        pass
    blur = ndi.median_filter(neg, footprint=fp, mode="reflect", cval=0.0)
    sums = cp.apply_along_axis(lambda a: a.sum(), axis, blur)
    line = sums.argmax(0)
    wsz = max(windowsize, margin)

    # Grab a +/- WINDOW-SIZE strip for evaluation. We've already cropped out the margins.
    if orientation == Orientation.HORIZONTAL:
        cropped = pos[max(line - wsz, 0) : min(line + wsz, height)]  # noqa: E203
    else:
        cropped = pos[
            :,
            max(line - wsz, 0) : min(line + wsz, width),  # noqa: E203
        ]

    # default footprint = scipy.ndimage.generate_binary_structure(image.ndim, 1)
    # array([[False, True, False],
    #        [ True, True,  True],
    #        [False, True, False]])
    # see https://github.com/scipy/scipy/issues/13991 : (grey_)dilation is faster than binary_dilation!

    edges = ndi.grey_dilation(
        canny(
            cropped,
            sigma=2.0,
            low_threshold=thresh[0],
            high_threshold=thresh[1],
            mode="reflect",
        ),
        footprint=ndi.generate_binary_structure(cropped.ndim, 1),
    )

    line_length = int(height * 0.25)
    lines = probabilistic_hough_line(
        get_cpu_image(edges), line_length=line_length, line_gap=2, theta=theta
    )

    # TODO: try OpenCV HoughLinesP
    # img_cv2 = cv_cuda_gpumat_from_cp_array(edges_u8)
    # print(edges_u8)
    # print(edges_u8.__cuda_array_interface__['data'][0])
    # print(img_cv2.download())
    # print(img_cv2.cudaPtr())
    # lines = cv2.HoughLinesP(img_cv2.download(),
    #                       1, #rho
    #                       theta,
    #                       10, #threshold
    #                       line_length, #minLineLength
    #                       2, #maxLineGap
    #                       )

    angles = []
    for (x0, y0), (x1, y1) in lines:
        # Ensure line is moving rightwards/upwards
        if orientation == Orientation.HORIZONTAL:
            k = 1 if x1 > x0 else -1
            offset = 0
        else:
            k = 1 if y1 > y0 else -1

            offset = 90
        angles.append(
            (
                orientation,
                offset - np.rad2deg(math.atan2(k * (y1 - y0), k * (x1 - x0))),
                x0,
                y0,
                x1,
                y1,
            )
        )
    return (angles, edges)


def analyse_image(
    image: tuple[
        Path,  # file
        tuple[int, int] | None,  # (pagenum, xref) if multi-image file
        tuple[int, int] | None,  # (xres, yres) dpi
        np.ndarray,  # image
    ],
):
    """Analyses an image for deskewing, potentially using a CUDA backend."""
    if "cuda" in dir(cp):  # pragma: no cover
        cp.cuda.set_pinned_memory_allocator(None)
    page = get_gpu_image_maybe(image[3])

    f = image[0]
    filename = f.name
    pagenum = float(f"{image[1][0]}.{image[1][1]}") if image[1] else 0.0

    global windowsize
    if "windowsize" not in globals():
        windowsize = 150

    if page.ndim > 2:
        logger.debug(f"{filename} p{pagenum} - multichannel, converting to greyscale")
        # discard any alpha channel
        page = rgb2gray(page[:, :, :3])
    pageh, pagew = page.shape

    # Remove the margins, which are often dirty due to skewing
    # 0.33" of an 8.5" page is approximately 1/25th
    pos = crop(page, pagew // 25)

    if is_low_contrast(pos):
        logger.debug(f"{filename} p{pagenum} - low contrast - blank page?")
        return (f, pagenum, "", "", pagew, pageh)

    thr = 255 * 0.8
    p = np.clip(pos, 0, thr)
    neg = thr - p
    if common.debug:
        dimg = get_cpu_image(neg)
        iio.imwrite(
            f"{common.debugpath}/{filename}_{pagenum}_neg.tiff",
            dimg,
        )

    # TODO: sanity check, _hough_angles should be under 100? 500?
    h_angles, h_edges = _hough_angles(pos, neg, Orientation.HORIZONTAL)
    v_angles, v_edges = _hough_angles(pos, neg, Orientation.VERTICAL)
    angles = h_angles + v_angles

    if len(angles) == 0:
        logger.debug(f"{filename} p{pagenum} - no lines found, changing threshold")
        if common.debug:
            dhedges = get_cpu_image(img_as_ubyte(_greyf(h_edges)))
            dvedges = get_cpu_image(img_as_ubyte(_greyf(v_edges)))
            iio.imwrite(
                f"{common.debugpath}/{filename}_{pagenum}_no_hlines.png",
                dhedges,
            )
            iio.imwrite(
                f"{common.debugpath}/{filename}_{pagenum}_no_vlines.png",
                dvedges,
            )
        h_angles, h_edges = _hough_angles(
            pos, neg, Orientation.HORIZONTAL, thresh=(100, 255)
        )
        v_angles, v_edges = _hough_angles(
            pos, neg, Orientation.VERTICAL, thresh=(100, 255)
        )
        angles = h_angles + v_angles
    else:
        if common.debug:
            dhedges = get_cpu_image(h_edges)
            dvedges = get_cpu_image(v_edges)
            iio.imwrite(
                f"{common.debugpath}/{filename}_{pagenum}_simple_dilation_h_edges.png",
                dhedges,
            )
            iio.imwrite(
                f"{common.debugpath}/{filename}_{pagenum}_simple_dilation_v_edges.png",
                dvedges,
            )

    if len(angles) > 0:
        angle = np.median([x[1] for x in angles])
        logger.debug(f"{filename} p{pagenum} - Hough angle median: {angle}°")
        if common.debug:
            hs = 0
            vs = 0
            h_edges_grey = _greyf(h_edges)
            v_edges_grey = _greyf(v_edges)
            for result in angles:
                orn, _, x0, y0, x1, y1 = result
                rr, cc, val = line_aa(c0=x0, r0=y0, c1=x1, r1=y1)
                if orn == Orientation.HORIZONTAL:
                    hs += 1
                    for k, v in enumerate(val):
                        h_edges_grey[rr[k], cc[k]] = (1 - v) * h_edges_grey[
                            rr[k], cc[k]
                        ] + v
                else:  # V
                    vs += 1
                    for k, v in enumerate(val):
                        v_edges_grey[rr[k], cc[k]] = (1 - v) * v_edges_grey[
                            rr[k], cc[k]
                        ] + v
            if hs > 0:
                dhg = get_cpu_image(img_as_ubyte(h_edges_grey))
                iio.imwrite(f"{common.debugpath}/{filename}_{pagenum}_hlines.png", dhg)
            if vs > 0:
                vhg = get_cpu_image(img_as_ubyte(v_edges_grey))
                iio.imwrite(f"{common.debugpath}/{filename}_{pagenum}_vlines.png", vhg)
        # end if common.debug
        angles = [x[1] for x in angles]

    else:
        logger.debug(f"{filename} p{pagenum} - Failed simple Hough, dilating...")

        # We didn't find a good feature at the H or V sum peaks.
        # Let's brutally dilate everything and look for a vertical margin!

        height, width = neg.shape
        if height * width > int(1e6):
            # note: small will be float64 here, so, slower on GPU
            small = downscale_local_mean(neg, (2, 2))
        else:
            small = neg

        # performance wise, binary_dilation was incredibly bad.
        # see   https://github.com/spinalcordtoolbox/spinalcordtoolbox/issues/4020#issuecomment-1409335950
        #   and https://scikit-image.org/docs/stable/auto_examples/applications/plot_rank_filters.html
        #   and https://github.com/scikit-image/scikit-image/issues/1190#issuecomment-57939688
        #       (ndimage and opencv faster than skimage.filter.rank.maximum)
        #   and the discussion here, decomposition is automatic for squares or rectangles:
        #       https://github.com/scipy/scipy/issues/13991#issuecomment-839954147
        #       https://github.com/scipy/scipy/blob/main/scipy/ndimage/_filters.py#L1723-L1730
        # pick a threshold for dilation
        thresh = threshold_otsu(small)
        dilated = ndi.grey_dilation(small > thresh, footprint=cp.ones((60, 60)))
        edges = canny(dilated, sigma=3.0)
        if common.debug:
            edges_grey = _greyf(edges)

        npedges = get_cpu_image(edges)
        lines = probabilistic_hough_line(
            npedges,
            line_length=int(pageh * 0.04),
            line_gap=6,
            theta=hough_theta_hv,
        )

        for (x_0, y_0), (x_1, y_1) in lines:
            if abs(x_1 - x_0) > abs(y_1 - y_0):
                # angle is <= π/4 from horizontal or vertical
                _, x0, y0, x1, y1 = Orientation.HORIZONTAL, x_0, y_0, x_1, y_1
            else:
                _, x0, y0, x1, y1 = Orientation.VERTICAL, y_0, -x_0, y_1, -x_1
            # flip angle so that X delta is positive (East quadrants).
            k = 1 if x1 > x0 else -1
            a = np.rad2deg(math.atan2(k * (y1 - y0), k * (x1 - x0)))

            # Zero angles are suspicious -- could be a cropping margin.
            # If not, they don't add information anyway.
            if a != 0:
                angles.append(-a)
            if common.debug:
                rr, cc, val = line_aa(c0=x_0, r0=y_0, c1=x_1, r1=y_1)
                for k, v in enumerate(val):
                    edges_grey[rr[k], cc[k]] = (1 - v) * edges_grey[rr[k], cc[k]] + v

        if angles:
            angle = np.median(angles)
            logger.debug(
                f"{filename} p{pagenum} - dilated Hough angle median: {angle}°"
            )
            if common.debug:
                dd = get_cpu_image(img_as_bool(dilated))
                deg = get_cpu_image(img_as_ubyte(edges_grey))
                iio.imwrite(
                    f"{common.debugpath}/{filename}_{pagenum}_dilated.png",
                    dd,
                )
                iio.imwrite(
                    f"{common.debugpath}/{filename}_{pagenum}_{angle}_lines.png",
                    deg,
                )
        else:
            angle = None
            logger.debug(f"{filename} p{pagenum} - Failed dilated Hough, giving up")
            if common.debug:
                dd = get_cpu_image(img_as_bool(dilated))
                deg = get_cpu_image(img_as_ubyte(edges_grey))
                de = get_cpu_image(img_as_bool(edges))
                iio.imwrite(f"{common.debugpath}/{filename}_{pagenum}_dilated.png", dd)
                iio.imwrite(
                    f"{common.debugpath}/{filename}_{pagenum}_dilate_edges_grey.png",
                    deg,
                )
                iio.imwrite(
                    f"{common.debugpath}/{filename}_{pagenum}_dilate_edges.png", de
                )

    return (
        f,
        pagenum,
        angle if angle is not None else "",
        np.var(angles) if angles else "",
        pagew,
        pageh,
    )


# NOTE: _init_worker must be in the same module as analyse_files (i.e. the call to the initializer), not sure why
def _init_worker(common_arg: CommonArgs):
    #    signal.signal(signal.SIGINT, signal.SIG_IGN)
    global common, logger
    common = common_arg
    logger = common_arg.logger


def analyse_files(files: list[Path], common: CommonArgs) -> Path:
    results = []

    with Pool(
        processes=common.workers, initializer=_init_worker, initargs=(common,)
    ) as p:
        try:
            with tqdm(
                total=get_num_images(files),
                disable=(common.loglevel != "WARNING"),
                unit="pg",
                desc="Analysis: ",
            ) as pbar:
                for i, result in enumerate(
                    p.imap_unordered(
                        analyse_image, get_images(files), min(common.workers, 8)
                    )
                ):
                    pbar.update()
                    bisect.insort(results, result, key=lambda x: x[1])
        except NotImplementedError as e:
            logger.error(e)
            abort(p)
            return None
        except KeyboardInterrupt:
            import sys

            print("Caught KeyboardInterrupt, terminating workers...", file=sys.stderr)
            abort(p)
            return None
        finally:
            # https://coverage.readthedocs.io/en/7.8.0/subprocess.html#using-multiprocessing
            p.close()
            p.join()

    # TODO: don't clobber existing results file?
    resultspath = Path(common.outpath, "results.csv")
    with open(resultspath, "w", newline="") as csvfile:
        writer = csv.writer(csvfile, quoting=csv.QUOTE_NONNUMERIC)
        writer.writerow(
            [
                "Input File",
                "Page Number",
                "Computed angle",
                "Variance of computed angles",
                "Image width (px)",
                "Image height (px)",
                # "X pixels per inch",
                # "Y pixels per inch",
            ]
        )
        writer.writerows(results)
    return resultspath



import pytest

from hough import (
    main,
)


def test_just_histogram():
    with pytest.raises(SystemExit) as e:
        main(["histogram", "--results", "newman-results.csv"])
    assert e.value.code == 0


def test_blank_histogram():
    with pytest.raises(SystemExit) as e:
        main(["histogram", "--results", "no-angles.csv"])
    assert e.value.code == 0


def test_invalid_csv_histogram():
    with pytest.raises(SystemExit) as e:
        main(["histogram", "--results", "../README.md"])
    assert e.value.code == 0


def test_null_histogram():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--histogram", "white.jpg"])
    assert e.value.code == 0


def test_histogram_verbose():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--histogram", "--verbose", "av-000.tif"])
    assert e.value.code == 0


def test_process_histogram_exception(mocker):
    mocker.patch("hough.stats._do_histogram", side_effect=RuntimeError("Boom"))
    with pytest.raises(SystemExit) as e:
        main(["process", "--histogram", "av-000.tif"])
    assert e.value.code == 1


# note: going to get DeprecationWarnings, cannot resolve yet
#   see https://github.com/swig/swig/issues/2881#issuecomment-2332652634



from importlib import metadata

import toml

from .utils import (
    CommonArgs,
    abort,
    load_csv,
    logit,
    start_pbar,
    get_cpu_image,
    get_gpu_image_maybe,
)
from .multifiles import get_images, get_num_images
from .stats import histogram
from .analyse import analyse_files, analyse_image
from .rotate import rotate_files
from .main import main


try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = toml.load("pyproject.toml")["tool"]["poetry"]["version"] + "dev"

__all__ = [
    "CommonArgs",
    "abort",
    "load_csv",
    "logit",
    "start_pbar",
    "get_cpu_image",
    "get_gpu_image_maybe",
    "get_images",
    "get_num_images",
    "histogram",
    "analyse_files",
    "analyse_image",
    "rotate_files",
    "main",
]



"""
Worker functions for a parallelizable deskewer.
"""

import signal
from multiprocessing import Queue
from multiprocessing.pool import Pool
from pathlib import Path

from loguru import logger

from . import start_pbar, get_cpu_image, get_gpu_image_maybe

try:  # pragma: no cover
    import cupy as cp
    import cupyx.scipy.ndimage as ndi
except ModuleNotFoundError:  # pragma: no cover
    import numpy as cp
    import scipy.ndimage as ndi

import filetype
import fitz
import imageio.v3 as iio
import numpy as np

from .utils import CommonArgs, abort, load_csv


# TODO: provide order override?
def rotate_ndarray(cpuimg: np.ndarray, angle: float, order=0) -> np.ndarray:
    gpuimg = get_gpu_image_maybe(cpuimg, dtype=cp.uint8)
    rotatedimg = ndi.rotate(gpuimg, -angle, mode="nearest", reshape=False, order=order)
    return get_cpu_image(rotatedimg)


def rotate(imagelist: list[tuple]) -> int:
    """Actually rotates a single file of 1+ images.
    Rotated file has the same name and is placed in common.outpath.
    """
    guess = filetype.guess(imagelist[0][0])
    if guess and guess.mime == "application/pdf":
        newdoc = fitz.open()
    else:
        newdoc = None
    filename = Path(imagelist[0][0])
    basename = filename.name
    kind = filetype.guess(filename)
    logger = common.logger

    for image in imagelist:
        page = int(image[1]) if image[1] else ""
        angle = float(image[2]) if image[2] else 0.0
        if angle == 0.0:
            continue
        if not page:
            # single-image file, not a container
            logger.info(f"Rotating {filename}...")
            cpuimg = iio.imread(image[0])
            fixed = rotate_ndarray(cpuimg, angle)
            iio.imwrite(f"{common.outpath}/{basename}", fixed)
            queue.put(1)
        else:
            if kind and kind.mime == "application/pdf":
                doc = fitz.open(image[0])
                imagelist = doc.get_page_images(page - 1)
                # TODO: Correctly deal with multiple images on a page
                for item in imagelist:
                    xref = item[0]
                    smask = item[1]
                    if smask == 0:
                        imgdict = doc.extract_image(xref)
                        logger.info(
                            f"Rotating {filename} - page {page} - xref {xref}..."
                        )
                        try:
                            cpuimg = iio.imread(imgdict["image"])
                            fixed = rotate_ndarray(cpuimg, angle)
                            imgext = imgdict["ext"]
                            imgbytes = iio.imwrite(
                                "<bytes>", fixed, extension=("." + imgext)
                            )
                            imgdoc = fitz.open(stream=imgbytes, filetype=imgext)
                            rect = imgdoc[0].rect
                            pdfbytes = imgdoc.convert_to_pdf()
                            imgdoc.close()
                            img_pdf = fitz.open("pdf", pdfbytes)
                            page = newdoc.new_page(width=rect.width, height=rect.height)
                            page.show_pdf_page(rect, img_pdf, 0)
                        except ValueError as e:
                            logger.error(
                                f"Skipping rotating {filename} - page {page} - xref {xref}: {e}"
                            )
                    else:
                        logger.error(
                            f"Skipping process {filename} - page {page} - image {xref} (smask=={smask})"
                        )
            # TODO: deal with other multi-image formats
            else:
                logger.error(
                    f"Skipping file {filename} - unknown multi-page file format"
                )
        queue.put(1)
    if newdoc:
        logger.info(f"Saving {common.outpath}/{basename}...")
        newdoc.save(f"{common.outpath}/{basename}")


# NOTE: _init_worker must be in the same module as analyse_files (i.e. the call to the initializer), not sure why
def _init_rotator(common_arg: CommonArgs, q: Queue):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    global common, logger, queue
    common = common_arg
    logger = common_arg.logger
    queue = q


def rotate_files(resultsfile: Path, common: CommonArgs) -> int:
    """Rotate all of the files mentioned in the results.csv resultsfile."""
    # TODO: further parallelise this by splitting multi-image files first?

    # load a previously generated results.csv file
    results = load_csv(resultsfile)
    if len(results) == 0:
        logger.error("Nothing to do!")
        return 1

    # create dict of lists, each key is a file and each value a list of tuple-images
    dictresults = {}
    for result in results:
        for image in result:
            dictresults.setdefault(image[0], []).append(image)
    sortedresults = [
        sorted(v, key=lambda x: int(x[1]) if x[1] else 0)
        for k, v in dictresults.items()
    ]

    try:
        q, proc = start_pbar(
            len(results),
            (common.loglevel != "WARNING"),
            "pg",
            "Rotation: ",
        )
        with Pool(
            processes=common.workers,
            initializer=_init_rotator,
            initargs=(common, q),
        ) as p:
            for i, result in enumerate(
                p.imap_unordered(rotate, sortedresults, min(common.workers, 8))
            ):
                pass
            # see https://coverage.readthedocs.io/en/7.8.0/subprocess.html#using-multiprocessing
            p.close()
            p.join()
        q.put(None)
        proc.join()
        return 0

    except KeyboardInterrupt:
        import sys

        print("Caught KeyboardInterrupt, terminating workers...", file=sys.stderr)
        abort(p, q, proc)
        return 1



import os
import sys
from multiprocessing import Pool
from pathlib import Path
from shutil import rmtree

import pytest

from hough import (
    __version__,
    main,
    abort,
    start_pbar,
)


def test_version():
    assert __version__


def test_abort():
    q, proc = start_pbar(1, True, "tests", "Testing: ")
    with Pool(processes=2) as p:
        abort(p, q, proc)
    with Pool(processes=2) as p:
        abort(p, 123, "abc")
    with Pool(processes=2) as p:
        abort(p, None, None)
    abort(123, 456, "abc")


def test_no_valid_files():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "a", "b", "c"])
    assert e.value.code == 1


@pytest.mark.parametrize("outpath", ["t"])
def test_multifile_debug(sampledir):
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--out=t", "--debug", "av-000.tif", "av-001.jpg"])
    assert e.value.code == 0
    assert Path("t/debug").exists()
    assert Path("t/hough.log").exists()
    # TODO: assert expected output appears under debug/DATE/whatever


@pytest.mark.parametrize("outpath", ["t"])
def test_reuse_outpath():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--out=t", "--debug", "binder.png"])
    assert e.value.code == 0
    assert Path("t/hough.log").exists()
    assert Path("t/results.csv").exists()
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--out=t", "--debug", "rhombus.jpg"])
    assert e.value.code == 0
    assert Path("t/hough.log").exists()
    assert Path("t/results.csv").exists()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="perm error in CI")
def test_broken_debug():
    Path("t").mkdir()
    open("t/debug", "a").close()
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--out=t", "--debug", "deep_hurting.png"])
    assert e.value.code == 1
    rmtree("t")


def test_no_args():
    with pytest.raises(SystemExit) as e:
        main([])
    assert e.value.code is None


@pytest.mark.skipif(sys.platform.startswith("win"), reason="perm error in CI")
def test_analyse_broken_out():
    open("t", "a").close()
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--out=t", "white.jpg"])
    assert e.value.code == 1
    os.remove("t")


@pytest.mark.skipif(sys.platform.startswith("win"), reason="perm error in CI")
@pytest.mark.parametrize("outpath", ["t"])
def test_rotate_broken_out():
    open("t", "a").close()
    with pytest.raises(SystemExit) as e:
        main(["rotate", "--results=t"])
    assert e.value.code == 1
    Path("t").unlink()


# note: going to get DeprecationWarnings, cannot resolve yet
#   see https://github.com/swig/swig/issues/2881#issuecomment-2332652634



import sys
from pathlib import Path

import pytest

from hough import (
    main,
)


def test_rotate_from_results():
    with pytest.raises(SystemExit) as e:
        main(["rotate", "--out=out", "--results=newman-results.csv"]) == 0
    assert e.value.code == 0
    assert Path("out/Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf").exists()


def test_rotate_smask_pdf():
    with pytest.raises(SystemExit) as e:
        main(["rotate", "--results=smask.csv"])
    assert e.value.code == 0


@pytest.mark.skipif(sys.platform.startswith("win"), reason="perm error in CI")
@pytest.mark.parametrize("outpath", ["t"])
def test_rotate_broken_out():
    open("t", "a").close()
    with pytest.raises(SystemExit) as e:
        main(["rotate", "--results=t"])
    assert e.value.code == 1
    Path("t").unlink()


def test_rotate_nothing_two():
    with pytest.raises(SystemExit) as e:
        main(["rotate", "nothing.at.all"])
    assert e.value.code == 1


# note: going to get DeprecationWarnings, cannot resolve yet
#   see https://github.com/swig/swig/issues/2881#issuecomment-2332652634



# -*- coding: utf-8 -*-
# vim: ts=4:sw=4:et:tw=88:nowrap

import csv
import datetime
import functools
import sys
import time
import traceback
from dataclasses import dataclass
from functools import cached_property
from multiprocessing import get_context, Process, Queue, Manager
from multiprocessing.pool import Pool
from pathlib import Path
from tqdm import tqdm

from cyclopts import Parameter
from loguru import logger

try:  # pragma: no cover
    import cupy as cp
except ModuleNotFoundError:  # pragma: no cover
    import numpy as cp
import numpy as np


def _setup_loguru(log_level, outpath):
    spawn_context = get_context("spawn")
    fmt = "{time:YYYY-MM-DD HH:mm:ss,SSS} {level:8} {process.name:>18}-{process.id:<7} {message}"
    # reattach stderr sink with queueing to ensure inheritability
    logger.remove()
    logger.add(
        sys.stderr,
        context=spawn_context,
        enqueue=True,
        level=log_level,
        format="{message}",
    )
    # TODO: do we still want this?
    logger.add(
        Path(outpath, "hough.log"),
        context=spawn_context,
        enqueue=True,
        level="DEBUG",
        format=fmt,
    )


def logit(func):
    """Initialise a function for logging"""

    @functools.wraps(func)
    def wrapper_logit(*args, **kwargs):
        # there must be an easier way
        if any(isinstance(val, CommonArgs) for val in args):
            for val in args:
                if isinstance(val, CommonArgs):
                    common = val
        else:
            common = CommonArgs()
            args = args + (common,)
        try:
            _setup_loguru(common.loglevel, common.outpath)
        except ValueError:
            print(
                "Unable to create output directory or logfile! Aborting.",
                file=sys.stderr,
            )
            exit(1)

        logger.info(
            f"=== Run started @ {datetime.datetime.now(datetime.UTC).isoformat()} ==="
        )
        value = func(*args, **kwargs)
        logger.info(
            f"=== Run ended  @ {datetime.datetime.now(datetime.UTC).isoformat()} ==="
        )
        return value

    return wrapper_logit


def get_cpu_image(img: cp.ndarray | np.ndarray) -> np.ndarray:  # pragma: no cover
    if "cuda" in dir(cp):
        return img.get()
    else:
        return img


def get_gpu_image_maybe(
    img: np.ndarray, dtype=np.float32
) -> cp.ndarray:  # pragma: no cover
    if "cuda" in dir(cp):
        return cp.asarray(img, dtype)
    else:
        return img


def _pbar_listener(q, total, disable, unit, desc):
    pbar = tqdm(total=total, disable=disable, unit=unit, desc=desc)
    for item in iter(q.get, None):
        pbar.update()


def start_pbar(total: int, disable: bool, unit: str, desc: str) -> (Queue, Process):
    q = Manager().Queue(-1)
    proc = Process(
        target=_pbar_listener,
        args=(q, total, disable, unit, desc),
    )
    proc.start()
    return q, proc


def abort(pool=None, log_queue=None, listener=None):
    try:
        if pool and isinstance(pool, Pool):
            pool.close()
            pool.terminate()
            pool.join()
            # this lets the producers drain their log queues
            time.sleep(0.1)
            print(
                f"=== Run killed @ {datetime.datetime.now(datetime.UTC).isoformat()} ===",
                file=sys.stderr,
            )
        if log_queue and listener:
            try:
                log_queue.put(None)
                listener.join()
            except Exception:
                pass
    except Exception:  # pragma: no cover
        print("Exception during abort:", file=sys.stderr)
        traceback.print_exc(file=sys.stderr)


def load_csv(f: Path):
    data = []
    try:
        with open(f, newline="") as csvfile:
            reader = csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)
            for row in reader:
                if row[0] == "Input File":
                    continue
                # for idx in [4, 5, 6, 7]:
                for idx in [4, 5]:
                    row[idx] = int(row[idx]) if row[idx] else ""
                row[2] = row[2] if row[2] else 0
                data.append([tuple(row)])
    except (OSError, ValueError) as e:
        logger.error(f"Unable to read results CSV {f}: {e}")
    return data


def validate_common(type_, value):
    if value and value.debug:
        # this will raise ValueError if there is an issue
        _ = value.debugpath


# name="*" flattens namespace so --analyse.debug becomes --debug
@Parameter(name="*", negative=(), validator=validate_common)
@dataclass
class CommonArgs:
    debug: bool = False
    "Save intermediate results in debug/ under out folder."

    verbose: bool = False
    "Print status messages instead of progress bar."

    out: Path = Path("out/TIMESTAMP")
    "Use the specified path for results and post-rotated files."

    workers: int = 4
    "Number of workers to run simultaneously."

    def __post_init__(self):
        if self.debug:
            self.loglevel = "DEBUG"
        elif self.verbose:
            self.loglevel = "INFO"
        else:
            self.loglevel = "WARNING"
        self.logger = logger

    @cached_property
    def outpath(self) -> Path:
        if "TIMESTAMP" in self.out.parts:
            self.out = Path(str(self.out).replace("TIMESTAMP", self.now))
        p = self.out.absolute()
        if not p.is_dir():
            try:
                p.mkdir(parents=True, exist_ok=True)
            except (FileExistsError, FileNotFoundError):
                raise ValueError(f"Unable to create output directory {p}")
        return p

    @cached_property
    def debugpath(self) -> Path:
        p = Path(self.outpath, "debug")
        if not p.is_dir():
            try:
                p.mkdir(parents=True, exist_ok=True)
            except (FileExistsError, FileNotFoundError):
                raise ValueError(f"Unable to create debug directory {p}")
        return p

    @cached_property
    def now(self) -> str:
        return datetime.datetime.now(datetime.UTC).strftime("%Y-%m-%dT%H%M%SZ")



from pathlib import Path

import pytest

from hough import (
    analyse_files,
    main,
    CommonArgs,
    load_csv,
)


def test_unorientable():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "unorientable.jpg"])
    assert e.value.code == 0


def test_unorientable_debug():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "--debug", "unorientable.jpg"])
    assert e.value.code == 0


def test_low_contrast():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "black.png"])
    assert e.value.code == 0


def test_white():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "white.jpg"])
    assert e.value.code == 0


def test_no_lines():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "rhombus.jpg"])
    assert e.value.code == 0


def test_small_blob():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "blob.png"])
    assert e.value.code == 0


def test_small_image():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "deep_hurting.png"])
    assert e.value.code == 0


def test_smask_pdf():
    with pytest.raises(SystemExit) as e:
        main(["process", "smask.pdf"])
    assert e.value.code == 0


def test_vert():
    resultspath = analyse_files([Path("av-000.tif")], CommonArgs())
    results = load_csv(resultspath)
    res = results[0][0]
    assert res[0] == "av-000.tif"
    assert res[1] == 0.0
    assert res[2] > -2 and res[2] <= 0.0
    assert res[3] < 1.0
    # assert(res[4] == 464)
    # assert(res[5] == 645)


def test_hv_fail():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "av-001.jpg"])
    assert e.value.code == 0


def test_unstraightenable():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "binder.png"])
    assert e.value.code == 0


@pytest.mark.parametrize("outpath", ["t"])
def test_rgb_pdf_4_workers():
    with pytest.raises(SystemExit) as e:
        main(
            [
                "analyse",
                "--debug",
                "--out=t",
                "--workers=4",
                "Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf",
            ]
        )
    assert e.value.code == 0
    assert Path("t/results.csv").exists()
    assert Path("t/hough.log").exists()
    assert Path("t/debug").exists()
    assert Path("t/debug").is_dir()


def test_form_pdf():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "i-9.pdf"])
    assert e.value.code == 0


def test_mixed_pdf():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "i-9-paper-version.pdf"])
    assert e.value.code == 0


def test_multipage_compressed_tiff():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "dg-nova-multipage.tif"]) == 0
    assert e.value.code == 0


def test_batch_tiff():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "av-multi-batch.tif"]) == 0
    assert e.value.code == 0


def test_unknown():
    with pytest.raises(SystemExit) as e:
        main(["analyse", "../README.md"])
    assert e.value.code == 1


# note: going to get DeprecationWarnings, cannot resolve yet
#   see https://github.com/swig/swig/issues/2881#issuecomment-2332652634



# -*- coding: utf-8 -*-
# vim: ts=4:sw=4:et:tw=88:nowrap

try:
    import cupy as cp
    import cv2

    # thank you, https://github.com/rapidsai/cucim/issues/329 !
    def cv_cuda_gpumat_from_cp_array(arr: cp.ndarray) -> cv2.cuda.GpuMat:
        assert len(arr.shape) in (
            2,
            3,
        ), "CuPy array must have 2 or 3 dimensions to be a valid GpuMat"
        type_map = {
            cp.dtype("uint8"): cv2.CV_8U,
            cp.dtype("int8"): cv2.CV_8S,
            cp.dtype("uint16"): cv2.CV_16U,
            cp.dtype("int16"): cv2.CV_16S,
            cp.dtype("int32"): cv2.CV_32S,
            cp.dtype("float32"): cv2.CV_32F,
            cp.dtype("float64"): cv2.CV_64F,
        }
        depth = type_map.get(arr.dtype)
        assert depth is not None, f"Unsupported CuPy array dtype {arr.dtype}"
        channels = 1 if len(arr.shape) == 2 else arr.shape[2]
        # equivalent to unexposed opencv C++ macro CV_MAKETYPE(depth,channels):
        # (depth&7) + ((channels - 1) << 3)
        mat_type = depth + ((channels - 1) << 3)
        # TODO: do we need [1::-1] here to invert the matrix?
        mat = cv2.cuda.createGpuMatFromCudaMemory(
            arr.__cuda_array_interface__["shape"][1::-1],
            mat_type,
            arr.__cuda_array_interface__["data"][0],
        )
        return mat

    def cp_array_from_cv_cuda_gpumat(mat: cv2.cuda.GpuMat) -> cp.ndarray:
        class CudaArrayInterface:
            def __init__(self, gpu_mat: cv2.cuda.GpuMat):
                w, h = gpu_mat.size()
                type_map = {
                    cv2.CV_8U: "|u1",
                    cv2.CV_8S: "|i1",
                    cv2.CV_16U: "<u2",
                    cv2.CV_16S: "<i2",
                    cv2.CV_32S: "<i4",
                    cv2.CV_32F: "<f4",
                    cv2.CV_64F: "<f8",
                }
                self.__cuda_array_interface__ = {
                    "version": 3,
                    "shape": (
                        (h, w, gpu_mat.channels()) if gpu_mat.channels() > 1 else (h, w)
                    ),
                    "typestr": type_map[gpu_mat.depth()],
                    "descr": [("", type_map[gpu_mat.depth()])],
                    "stream": 1,
                    "strides": (
                        (gpu_mat.step, gpu_mat.elemSize(), gpu_mat.elemSize1())
                        if gpu_mat.channels() > 1
                        else (gpu_mat.step, gpu_mat.elemSize())
                    ),
                    "data": (gpu_mat.cudaPtr(), False),
                }

        arr = cp.asarray(CudaArrayInterface(mat))
        return arr

except ModuleNotFoundError:
    pass



# -*- coding: utf-8 -*-
# vim: ts=4:sw=4:et:tw=88:nowrap

from collections.abc import Generator
from pathlib import Path

import filetype
import imageio.v3 as iio
import pymupdf
from loguru import logger
from numpy import ndarray


def get_pages(
    f: Path,
) -> list[tuple[Path, str, int]]:  # pragma: no cover because dead code
    """Returns the pages in the file, as a list of tuples of the form:
    [(filename, "mime/type", pagenum), ... ]
    """
    kind = filetype.guess(f)
    if not kind:
        # assume single-image file
        return [(f, kind.mime, None)]
    if kind.mime == "application/pdf":
        pdf = pymupdf.open(f)
        return [(f, kind.mime, x) for x in range(len(pdf))]
    elif kind.mime == "image/tiff":
        # at least 2 ways this could be a multi-image
        # ignoring volumetric tiffs for now
        # first try the simple way
        props = iio.improps(f, index=...)
        if props.n_images > 1:
            # retrieve with iio.imread(f, index=n)
            return [(f, kind.mime, x) for x in range(props.n_images)]
        elif props.is_batch:
            # degenerate scum. sadly, this requires reading the entire thing in
            count = iio.imread(f).shape[0]
            return [(f, "image/tiff-batch", x) for x in range(count)]
        else:
            return [(f, kind.mime, None)]
    # TODO: Add support for multi-page TIFFs here via
    #   https://imageio.readthedocs.io/en/stable/reference/userapi.html#reading-images
    #   https://imageio.readthedocs.io/en/stable/userapi.html#imageio.mimread
    else:
        # assume single-image file
        return [(f, kind.mime, None)]


def get_num_images(files: list[Path]) -> int:
    ctr = 0

    for f in files:
        kind = filetype.guess(f)
        if not kind:
            return 0
        if kind.mime == "application/pdf":
            pdf = pymupdf.open(f)
            for page in pdf.pages():
                for image in page.get_images():
                    ctr += 1
        elif kind.mime == "image/tiff":
            props = iio.improps(f, index=...)
            if props.n_images > 1:
                ctr += props.n_images
            else:
                img = iio.imread(f)
                if len(img.shape) == 3:
                    # batch-style, indexed from zero
                    ctr += img.shape[0] + 1
                else:
                    # assume single-image tiff
                    ctr += 1
        else:
            ctr += 1
    return ctr


def get_images(
    files: list[Path],
) -> Generator[
    tuple[
        Path,  # file
        tuple[int, int] | None,  # (pagenum, xref) if multi-image file
        tuple[int, int] | None,  # (xres, yres) dpi
        ndarray,  # image
    ]
]:
    """Generator to return an ndarray for each image in the list of files"""
    for f in files:
        # do not catch exceptions here, let them bubble up
        kind = filetype.guess(f)
        if kind and kind.mime == "application/pdf":
            logger.info(f"Reading PDF {f}...")
            pdf = pymupdf.open(f)
            for page in pdf.pages():
                for image in page.get_images():
                    if image[1] != 0:
                        logger.warning(
                            f"Skipping {f} page {page.number}, non-zero smask {image[1]}"
                        )
                    d = pdf.extract_image(image[0])
                    logger.info(
                        f"Reading PDF {f} page {page.number} image {image[0]}..."
                    )
                    # TODO: add colorspace info?
                    yield (
                        f,
                        (page.number + 1, image[0]),
                        (d["xres"], d["yres"]),
                        iio.imread(d["image"]),
                    )
        elif kind and kind.mime == "image/tiff":
            logger.info(f"Reading TIFF {f}...")
            # at least 2 ways this could be a multi-image
            # (ignoring volumetric tiffs for now)
            # first try the simple way
            props = iio.improps(f, index=...)
            if props.n_images > 1:
                # retrieve with iio.imread(f, index=n)
                logger.info(f"TIFF {f} is multi-image ({props.n_images} images)...")
                for n in range(props.n_images):
                    logger.info(f"Reading TIFF {f} image {n}...")
                    img = iio.imread(f, index=n)
                    yield (f, (n + 1, 0), img.shape, img)
            else:
                # batch marker is not indicative; guess based on dimensions
                logger.info(f"Reading possibly batch-style TIFF {f}...")
                img = iio.imread(f)
                if len(img.shape) == 3:
                    # degenerate scum. sadly, this requires reading the entire thing in
                    logger.info(f"Batch-style TIFF {f} has {img.shape[0]} images.")
                    for n in range(img.shape[0]):
                        yield (f, (n + 1, 0), img[n].shape, img[n])
                else:
                    # assume single-image tiff
                    yield (f, None, None, iio.imread(f))
        else:
            try:
                logger.info(f"Reading {f}...")
                yield (f, None, None, iio.imread(f))
            except OSError:
                raise NotImplementedError(f"Unknown file type: {f}")
